# =============================================================================
# Casanova Reader
# =============================================================================
#
# A fast but comfortable CSV reader based upon csv.reader to avoid dealing
# with csv.DictReader which is nice but very slow.
#
import csv
from collections.abc import Iterable
from io import IOBase
from operator import itemgetter

from casanova.defaults import DEFAULTS
from casanova.utils import ensure_open, suppress_BOM, size_of_row_in_file
from casanova.exceptions import EmptyFileError, MissingColumnError, NoHeadersError


def validate_multiplex_tuple(multiplex):
    return (
        isinstance(multiplex, tuple) and
        len(multiplex) in [2, 3] and
        all(isinstance(t, str) for t in multiplex)
    )


class DictLikeRow(object):
    __slots__ = ('__mapping', '__row')

    def __init__(self, mapping, row):
        self.__mapping = mapping
        self.__row = row

    def __getitem__(self, key):
        return self.__row[self.__mapping[key]]

    def __getattr__(self, key):
        return self.__getitem__(key)


class Headers(object):
    def __init__(self, fieldnames):
        self.__mapping = {h: i for i, h in enumerate(fieldnames)}

    def rename(self, old_name, new_name):
        if old_name == new_name:
            raise TypeError

        self.__mapping[new_name] = self[old_name]
        del self.__mapping[old_name]

    def __len__(self):
        return len(self.__mapping)

    def __getitem__(self, key):
        return self.__mapping[key]

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __contains__(self, key):
        return key in self.__mapping

    def __iter__(self):
        yield from sorted(self.__mapping.items(), key=itemgetter(1))

    def as_dict(self):
        return self.__mapping.copy()

    def get(self, key, default=None):
        return self.__mapping.get(key, default)

    def collect(self, keys):
        return [self[k] for k in keys]

    def wrap(self, row):
        return DictLikeRow(self.__mapping, row)

    def __repr__(self):
        class_name = self.__class__.__name__

        representation = '<' + class_name

        for h, i in self:
            if h.isidentifier():
                representation += ' %s=%s' % (h, i)
            else:
                representation += ' "%s"=%s' % (h, i)

        representation += '>'

        return representation


class Reader(object):
    namespace = 'casanova.reader'

    def __init__(self, input_file, no_headers=False, encoding='utf-8',
                 dialect=None, quotechar=None, delimiter=None, prebuffer_bytes=None,
                 total=None, multiplex=None, ignore_null_bytes=None):

        # Resolving global defaults
        if prebuffer_bytes is None:
            prebuffer_bytes = DEFAULTS['prebuffer_bytes']

        if ignore_null_bytes is None:
            ignore_null_bytes = DEFAULTS['ignore_null_bytes']

        if not isinstance(ignore_null_bytes, bool):
            raise TypeError('expecting a boolean as "ignore_null_bytes" kwarg')

        self.ignore_null_bytes = ignore_null_bytes

        # Detecting input type
        if isinstance(input_file, IOBase):
            input_type = 'file'

        elif isinstance(input_file, str):
            input_type = 'path'
            input_file = ensure_open(input_file, encoding=encoding)

        elif isinstance(input_file, Iterable):
            input_type = 'iterable'
            input_file = iter(input_file)

        else:
            raise TypeError('expecting a file, a path or an iterable of rows')

        if multiplex is not None and not validate_multiplex_tuple(multiplex):
            raise TypeError('`multiplex` should be a 2-tuple or 3-tuple containing the column to split, the split character and optionally a new name for the column')

        reader_kwargs = {}

        if dialect is not None:
            reader_kwargs['dialect'] = dialect
        if quotechar is not None:
            reader_kwargs['quotechar'] = quotechar
        if delimiter is not None:
            reader_kwargs['delimiter'] = delimiter

        self.input_type = input_type
        self.input_file = input_file

        if self.input_type == 'iterable':
            self.reader = self.input_file
        else:
            if ignore_null_bytes:
                input_file = (item.replace('\0', '') for item in input_file)
            self.reader = csv.reader(input_file, **reader_kwargs)

        self.buffered_rows = []
        self.was_completely_buffered = False
        self.total = total
        self.headers = None
        self.expected_row_length = None

        # Reading headers
        if no_headers:
            try:
                self.buffered_rows.append(next(self.reader))
            except StopIteration:
                raise EmptyFileError

            self.expected_row_length = len(self.buffered_rows[0])
        else:
            try:
                fieldnames = next(self.reader)

                if fieldnames:
                    fieldnames[0] = suppress_BOM(fieldnames[0])

            except StopIteration:
                raise EmptyFileError

            self.headers = Headers(fieldnames)

        # Multiplexing
        if multiplex is not None:
            multiplex_column = multiplex[0]
            split_char = multiplex[1]

            if multiplex_column not in self.headers:
                raise MissingColumnError(multiplex_column)

            multiplex_pos = self.headers[multiplex_column]

            # New col
            if len(multiplex) == 3:
                self.headers.rename(multiplex_column, multiplex[2])

            original_reader = self.reader

            def reader_wrapper():
                for row in original_reader:
                    cell = row[multiplex_pos]

                    if not cell or split_char not in cell:
                        yield row

                    else:
                        for value in cell.split(split_char):
                            copy = list(row)
                            copy[multiplex_pos] = value
                            yield copy

            self.reader = reader_wrapper()

        # Prebuffering
        if prebuffer_bytes is not None and self.total is None:
            if not isinstance(prebuffer_bytes, int) or prebuffer_bytes < 1:
                raise TypeError('expecting a positive integer as "prebuffer_bytes" kwarg')

            buffered_bytes = 0

            while buffered_bytes < prebuffer_bytes:
                row = next(self.reader, None)

                if row is None:
                    self.was_completely_buffered = True
                    self.total = len(self.buffered_rows)
                    break

                buffered_bytes += size_of_row_in_file(row)
                self.buffered_rows.append(row)

        self.buffered_rows.reverse()

        # Iteration state
        self.prelude_rows = None
        self.row_filter = None
        self.current_row_index = -1

    def __repr__(self):
        columns_info = ' '.join('%s=%s' % t for t in self.headers)

        return '<%s %s>' % (self.namespace, columns_info)

    @property
    def fieldnames(self):
        if self.headers is None:
            return None

        return [k for k, v in self.headers]

    @property
    def row_len(self):
        if self.expected_row_length is not None:
            return self.expected_row_length

        return len(self.headers)

    def rows(self):

        def chained():
            if self.prelude_rows is not None:
                for row in self.prelude_rows:
                    self.current_row_index += 1
                    yield row

            while self.buffered_rows:
                self.current_row_index += 1
                yield self.buffered_rows.pop()

            for row in self.reader:
                self.current_row_index += 1
                yield row

        if self.row_filter is None:
            yield from chained()

        for row in chained():
            if self.row_filter(self.current_row_index, row):
                yield row

    def enumerate(self):
        for row in self.rows():
            yield self.current_row_index, row

    def wrap(self, row):
        return self.headers.wrap(row)

    def __iter__(self):
        return self.rows()

    def __cells(self, column, with_rows=False):
        if not isinstance(column, int):
            if self.headers is None:
                raise NoHeadersError

            pos = self.headers.get(column)

            if pos is None:
                raise MissingColumnError(column)
        else:
            if column >= self.row_len:
                raise MissingColumnError

            pos = column

        if with_rows:
            def iterator():
                for row in self.rows():
                    yield row, row[pos]
        else:
            def iterator():
                for row in self.rows():
                    yield row[pos]

        return iterator()

    def cells(self, column, with_rows=False):
        if not isinstance(column, (str, int)):
            raise TypeError

        return self.__cells(column, with_rows=with_rows)

    def close(self):
        if self.input_type == 'file':
            self.input_file.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @classmethod
    def count(cls, input_file, max_rows=None, **kwargs):
        assert max_rows is None or max_rows > 0, '%s.count: expected max_rows to be `None` or > 0.' % cls.namespace

        n = 0

        with cls(input_file, **kwargs) as reader:
            for _ in reader:
                n += 1

                if max_rows is not None and n > max_rows:
                    return None

        return n



# =============================================================================
# Casanova Named Record Helper
# =============================================================================
#
# CSV-aware improvement over python's namedtuple.
#
from json import dumps
from collections import OrderedDict
from collections.abc import Iterable

from casanova._namedtuple import future_namedtuple


def namedrecord(name, fields, boolean=None, plural=None, json=None, defaults=None):
    mapping = {k: i for i, k in enumerate(fields)}
    mask = []

    for k in fields:
        if boolean and k in boolean:
            mask.append(1)
        elif plural and k in plural:
            mask.append(2)
        elif json and k in json:
            mask.append(3)
        else:
            mask.append(0)

    def cast_for_csv(i, v, plural_separator='|'):
        if v is None:
            return None

        m = mask[i]

        if m == 0:
            return v

        if m == 1:
            return 'true' if v else 'false'

        if m == 2:
            assert isinstance(v, Iterable)
            return plural_separator.join(str(i) for i in v)

        if m == 3:
            return dumps(v, ensure_ascii=False)

        raise TypeError

    class Record(future_namedtuple(name, fields, defaults=defaults)):
        def __getitem__(self, key):
            if isinstance(key, str):
                idx = mapping.get(key)

                if idx is None:
                    raise KeyError

                return super().__getitem__(idx)

            return super().__getitem__(key)

        def get(self, key, default=None):
            try:
                return self.__getitem__(key)
            except (IndexError, KeyError):
                return default

        def as_csv_row(self, plural_separator='|'):
            row = list(
                cast_for_csv(i, v, plural_separator=plural_separator)
                for i, v in enumerate(self)
            )

            return row

        def as_csv_dict_row(self, plural_separator='|'):
            row = OrderedDict(
                (fields[i], cast_for_csv(i, v, plural_separator=plural_separator))
                for i, v in enumerate(self)
            )

            return row

        def as_dict(self):
            return {fields[i]: v for i, v in enumerate(self)}

    Record.__name__ = name
    Record.fieldnames = list(fields)

    return Record



# =============================================================================
# Casanova Reverse Reader
# =============================================================================
#
# A reader reading the file backwards in order to read last lines in constant
# time. This is sometimes useful to be able to resume some computations
# where they were left off.
#
import csv
from io import DEFAULT_BUFFER_SIZE
from file_read_backwards.file_read_backwards import FileReadBackwardsIterator
from ebbe import with_is_last

from casanova.reader import Reader
from casanova.utils import ensure_open
from casanova.exceptions import EmptyFileError, MissingColumnError

END_OF_FILE = object()


class Batch(object):
    __slots__ = ('value', 'finished', 'cursor', 'rows')

    def __init__(self, value, finished=False, cursor=None, rows=None):
        self.value = value
        self.finished = finished
        self.cursor = cursor
        self.rows = rows or []

    def __eq__(self, other):
        return (
            self.value == other.value and
            self.finished == other.finished and
            self.cursor == other.cursor and
            self.rows == other.rows
        )

    def __iter__(self):
        return iter(self.rows)

    def collect(self, pos):
        return set(row[pos] for row in self)

    def __repr__(self):
        class_name = self.__class__.__name__

        return (
            '<%(class_name)s value=%(value)s finished=%(finished)s cursor=%(cursor)s rows=%(rows)i>'
        ) % {
            'class_name': class_name,
            'value': self.value,
            'finished': self.finished,
            'cursor': self.cursor,
            'rows': len(self.rows)
        }


class ReverseReader(Reader):
    namespace = 'casanova.reverse_reader'

    def __init__(self, input_file, quotechar=None, **kwargs):
        super().__init__(input_file, quotechar=quotechar, **kwargs)
        quotechar = quotechar or '"'

        self.backwards_file = ensure_open(self.input_file.name, mode='rb')

        backwards_iterator = FileReadBackwardsIterator(
            self.backwards_file,
            self.input_file.encoding,
            DEFAULT_BUFFER_SIZE
        )

        def correctly_escaped_backwards_iterator():
            acc = None

            for line in backwards_iterator:
                if acc is not None:
                    acc = line + '\n' + acc
                else:
                    acc = line

                if acc.count(quotechar) % 2 == 0:
                    if self.ignore_null_bytes:
                        acc = acc.replace('\0', '')
                    yield acc
                    acc = None

            if acc is not None:
                if self.ignore_null_bytes:
                    acc = acc.replace('\0', '')
                yield acc

        backwards_reader = csv.reader(correctly_escaped_backwards_iterator())

        def generator():
            for is_last, row in with_is_last(backwards_reader):
                if not is_last or self.fieldnames is None:
                    yield row

            self.close()

        self.reader = generator()

        if self.fieldnames is None:
            self.buffered_rows = []

    def close(self):
        super().close()
        self.backwards_file.close()

    @staticmethod
    def last_cell(input_file, column, **kwargs):
        with ReverseReader(input_file, **kwargs) as reader:
            record = next(reader.cells(column), END_OF_FILE)

            if record is END_OF_FILE:
                raise EmptyFileError

            return record

    @staticmethod
    def last_batch(input_file, batch_value, batch_cursor, end_symbol, **kwargs):
        with ReverseReader(input_file, **kwargs) as reader:
            batch = END_OF_FILE

            if batch_value not in reader.headers:
                raise MissingColumnError(batch_value)

            if batch_cursor not in reader.headers:
                raise MissingColumnError(batch_cursor)

            batch_value_pos = reader.headers[batch_value]
            batch_cursor_pos = reader.headers[batch_cursor]

            for row in reader:
                value = row[batch_value_pos]
                cursor = row[batch_cursor_pos]

                if batch is END_OF_FILE:
                    batch = Batch(value)

                if value != batch.value:
                    return batch

                if cursor == end_symbol:
                    batch.finished = True
                    return batch

                if cursor:
                    batch.cursor = cursor
                    return batch

                batch.rows.append(row)

            if batch is END_OF_FILE:
                raise EmptyFileError

            return batch



import sys
from operator import itemgetter
from keyword import iskeyword


def _tuplegetter(index, doc):
    return property(itemgetter(index), doc=doc)


# From: https://github.com/python/cpython/blob/master/Lib/collections/__init__.py
def future_namedtuple(typename, field_names, *, rename=False, defaults=None, module=None):
    """Returns a new subclass of tuple with named fields.
    >>> Point = namedtuple('Point', ['x', 'y'])
    >>> Point.__doc__                   # docstring for the new class
    'Point(x, y)'
    >>> p = Point(11, y=22)             # instantiate with positional args or keywords
    >>> p[0] + p[1]                     # indexable like a plain tuple
    33
    >>> x, y = p                        # unpack like a regular tuple
    >>> x, y
    (11, 22)
    >>> p.x + p.y                       # fields also accessible by name
    33
    >>> d = p._asdict()                 # convert to a dictionary
    >>> d['x']
    11
    >>> Point(**d)                      # convert from a dictionary
    Point(x=11, y=22)
    >>> p._replace(x=100)               # _replace() is like str.replace() but targets named fields
    Point(x=100, y=22)
    """

    # Validate the field names.  At the user's option, either generate an error
    # message or automatically replace the field name with a valid name.
    if isinstance(field_names, str):
        field_names = field_names.replace(',', ' ').split()
    field_names = list(map(str, field_names))
    typename = sys.intern(str(typename))

    if rename:
        seen = set()
        for index, name in enumerate(field_names):
            if (
                not name.isidentifier() or
                iskeyword(name) or
                name.startswith('_') or
                name in seen
            ):
                field_names[index] = f'_{index}'
            seen.add(name)

    for name in [typename] + field_names:
        if type(name) is not str:
            raise TypeError('Type names and field names must be strings')
        if not name.isidentifier():
            raise ValueError('Type names and field names must be valid '
                             f'identifiers: {name!r}')
        if iskeyword(name):
            raise ValueError('Type names and field names cannot be a '
                             f'keyword: {name!r}')

    seen = set()
    for name in field_names:
        if name.startswith('_') and not rename:
            raise ValueError('Field names cannot start with an underscore: '
                             f'{name!r}')
        if name in seen:
            raise ValueError(f'Encountered duplicate field name: {name!r}')
        seen.add(name)

    field_defaults = {}
    if defaults is not None:
        defaults = tuple(defaults)
        if len(defaults) > len(field_names):
            raise TypeError('Got more default values than field names')
        field_defaults = dict(reversed(list(zip(reversed(field_names),
                                                reversed(defaults)))))

    # Variables used in the methods and docstrings
    field_names = tuple(map(sys.intern, field_names))
    num_fields = len(field_names)
    arg_list = ', '.join(field_names)
    if num_fields == 1:
        arg_list += ','
    repr_fmt = '(' + ', '.join(f'{name}=%r' for name in field_names) + ')'
    tuple_new = tuple.__new__
    _dict, _tuple, _len, _map, _zip = dict, tuple, len, map, zip

    # Create all the named tuple methods to be added to the class namespace

    namespace = {
        '_tuple_new': tuple_new,
        '__builtins__': {},
        '__name__': f'namedtuple_{typename}',
    }
    code = f'lambda _cls, {arg_list}: _tuple_new(_cls, ({arg_list}))'
    __new__ = eval(code, namespace)
    __new__.__name__ = '__new__'
    __new__.__doc__ = f'Create new instance of {typename}({arg_list})'
    if defaults is not None:
        __new__.__defaults__ = defaults

    @classmethod
    def _make(cls, iterable):
        result = tuple_new(cls, iterable)
        if _len(result) != num_fields:
            raise TypeError(
                f'Expected {num_fields} arguments, got {len(result)}')
        return result

    _make.__func__.__doc__ = (f'Make a new {typename} object from a sequence '
                              'or iterable')

    def _replace(self, **kwds):
        result = self._make(_map(kwds.pop, field_names, self))
        if kwds:
            raise ValueError(f'Got unexpected field names: {list(kwds)!r}')
        return result

    _replace.__doc__ = (f'Return a new {typename} object replacing specified '
                        'fields with new values')

    def __repr__(self):
        'Return a nicely formatted representation string'
        return self.__class__.__name__ + repr_fmt % self

    def _asdict(self):
        'Return a new dict which maps field names to their values.'
        return _dict(_zip(self._fields, self))

    def __getnewargs__(self):
        'Return self as a plain tuple.  Used by copy and pickle.'
        return _tuple(self)

    # Modify function metadata to help with introspection and debugging
    for method in (
        __new__,
        _make.__func__,
        _replace,
        __repr__,
        _asdict,
        __getnewargs__,
    ):
        method.__qualname__ = f'{typename}.{method.__name__}'

    # Build-up the class namespace dictionary
    # and use type() to build the result class
    class_namespace = {
        '__doc__': f'{typename}({arg_list})',
        '__slots__': (),
        '_fields': field_names,
        '_field_defaults': field_defaults,
        '__new__': __new__,
        '_make': _make,
        '_replace': _replace,
        '__repr__': __repr__,
        '_asdict': _asdict,
        '__getnewargs__': __getnewargs__,
        '__match_args__': field_names,
    }
    for index, name in enumerate(field_names):
        doc = sys.intern(f'Alias for field number {index}')
        class_namespace[name] = _tuplegetter(index, doc)

    result = type(typename, (tuple,), class_namespace)

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in environments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython), or where the user has
    # specified a particular module.
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass
    if module is not None:
        result.__module__ = module

    return result
